% File : LF_ACDC_solve_V2_1.m
% By:  Silvano Casoria
% Version: 2.1
% Matlab : (R2016b)
% Description: Script file for the sequential DC and AC load flow
% method involving iteration between DC and AC solutions.
% -----------------------------------------------------------------------------

%%  Read parameters defined in the Simulink model mask

save('OriginalWorkSpace.mat'); % Save the original workspace, to reloaded at the end.
clear;

ModelName_Sequential_LF_Param = gcs;

% Set the original workspace if normal ending
Set_original = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Set_original'); 
flag_clear   = strcmp(Set_original,'on');  % Flag (1/0)

%  The AC-DC network model
ModelName_Network  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Model_network'); 

if  (exist(eval('ModelName_Network'),'file') ~= 4) 
    msg_AC  = sprintf('Network model ''%s'' not found.',ModelName_Network);
    errordlg(msg_AC,'File Error');
    fprintf(['Error --> ',msg_AC,'\n']);
    clear;
    load OriginalWorkSpace.mat;
    return;    % Exit
end

% First DC Link steady state model
 ModelName_LFDClink     = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'LFDClink');
  if  (exist(eval('ModelName_LFDClink'),'file') ~= 4) 
    msg2  = sprintf('HVDC Link: Steady-state model ''%s'' not found.',ModelName_LFDClink);
    errordlg(msg2,'File Error');
    fprintf(['Error --> ',msg2,'\n']);
    clear;
    load OriginalWorkSpace.mat;
    return            % Exit
  else
   open (ModelName_LFDClink);
  end
  
% Indicates the presence of a second DC link
second_link    = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'second_link'); 
two_links      = strcmp(second_link,'on');  % Flag (1/0) 

if (two_links)
% Second DC Link steady state model
  ModelName_LFDClink_B     = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'LFDClink_B');
  if  (exist(eval('ModelName_LFDClink_B'),'file') ~= 4) 
    msg2  = sprintf('Second HVDC Link: Steady-state model ''%s'' not found.',ModelName_LFDClink_B);
    errordlg(msg2,'File Error');
    fprintf(['Error --> ',msg2,'\n']);
    clear;
    load OriginalWorkSpace.mat;
    return            % Exit
  else
   open (ModelName_LFDClink_B);
  end
end

fprintf('\n////// Sequential DC and AC load flow solutions  //////\n');
fprintf('Note: Your Workspace was saved in the file ''OriginalWorkSpace.mat''.\n\n'); 
fprintf('--> Opened the network model and the DC link(s) steady-state model(s).\r...Solving ...\n');

open (ModelName_Network);

% === Link A 
% DateString = datestr(datetime('now'));   % Date & time of the calculation
DateString = datestr(clock);   % Date & time of the calculation  %??
ReportName   = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Report'); %  The results text file

% The load flow bus name
bus_recName  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'bus_rec');
bus_invName  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'bus_inv');

% The dynamic load name (note: include the its subsystem name)
Load_recName  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Load_rec');
Load_invName  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Load_inv');

fileID   = fopen(ReportName,'w');

if (two_links)
% === link B
ReportName_B   = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Report_B'); %  The results text file

bus_recName_B  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'bus_rec_B');
bus_invName_B  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'bus_inv_B');

% The dynamic load name (note: include the its subsystem name)
Load_recName_B  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Load_rec_B');
Load_invName_B  = get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'Load_inv_B');

fileID_B = fopen(ReportName_B,'w');
end

% Initialization :
AC_converge      = 1;    % Flags
Normal_end_B     = 1;
Normal_end       = 1;
AC_DC_converge   = 1;
AC_DC_converge_B = 1;

% Allowed number of iterations for DC load flow soultion
DC_max_iter = eval(get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'DC_max_iter'));
% Allowed number of AC-DC load flows
DC_AC_max   = eval(get_param([ModelName_Sequential_LF_Param,'/Sequential Load Flow Param'],'DC_AC_max'));

% === Link A
Eac_R     = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'Eac_R_INI'));
Eac_I     = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'Eac_I_INI'));
TAPINI_R  = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'TAP_R_INI'));
TAPINI_I  = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'TAP_I_INI'));
U1_R   = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'U1_R'));
UacNom_R = U1_R(1);
U1_I   = eval(get_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'U1_I'));
UacNom_I = U1_I(1);

if (two_links)
% === Link B
Eac_R_B     = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'Eac_R_INI'));
Eac_I_B     = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'Eac_I_INI'));
TAPINI_R_B  = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'TAP_R_INI'));
TAPINI_I_B  = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'TAP_I_INI'));
U1_R_B   = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'U1_R'));
UacNom_R_B = U1_R_B(1);
U1_I_B   = eval(get_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'U1_I'));
UacNom_I_B = U1_I_B(1);
end

%% 
% Determine the DC load flow solution of the 2 terminal HVDC link(s) using the 
% steady state model Simulink bloc.  

   for i=1:DC_AC_max
       
    % ==== Link A ===
    sim(ModelName_LFDClink);    % Execute the steady-state model of the DC link
      
    if (i == 1)
        fprintf(fileID,'/// Sequential DC and AC load flows solution report: %s ///\rNetwork model: %s\n',...
        ReportName,ModelName_Network);
        fprintf(fileID,'DC link steady-state model: %s\n',ModelName_LFDClink);
        fprintf(fileID,'Date: %s\n\n',DateString);
        fprintf(fileID,'______ Execute the DC load flow (Iteration No %d) ______\n',i);
        fprintf(fileID,'Assumed values from AC network: \n');
    else
       fprintf(fileID,'______ Execute the DC load flow (Iteration No %d) ______\n',i);
       fprintf(fileID,'Inputs from AC load flow: \n');
    end
    
    fprintf(fileID,'Rectifier AC voltage: Eac_R = %5.2f kV\n', Eac_R(end));
    fprintf(fileID,'Inverter  AC voltage: Eac_I = %5.2f kV\n', Eac_I(end));
     
    if (DC_iter(end) <= DC_max_iter)

        fprintf(fileID,'--> Converges after %d iteration(s):\n',DC_iter(end)- 0);
        
        fprintf(fileID,'Rectifier: Tap = %5.4f (pu), alpha = %4.2f (deg), Udi0 = %5.2f (kV)\n',...
          Tap_pu_R(end), alpha_R(end), Udi0_R(end));
        fprintf(fileID,'Inverter:  Tap = %5.4f (pu), gamma = %4.2f (deg), Udi0 = %5.2f (kV)\n\n',...
          Tap_pu_I(end), gamma_I(end), Udi0_I(end));
%    
        Prec=Pac_R(end);    % AC active power at rectifier (MW)
        Qrec=Qac_R(end);    % AC reactive power at rectifier (Mvar)
        Pinv=Pac_I(end);    % AC active power at inverter (MW)
        Qinv=Qac_I(end);    % AC reactive power at inverter (Mvar)
    
        % Write P, Q values, for the next AC load flow
        try
          set_param([ModelName_Network,'/',Load_recName],'ActiveReactivePowers', sprintf('[%g %g ]',Prec*1e6 , Qrec*1e6));
        catch
          msg_block = sprintf('HVDC Link: No block (PQ load) called ''%s'' could be found.',Load_recName);
          errordlg(msg_block,'Block Error');
          clear;
          load OriginalWorkSpace.mat;  % Load the original workspace
          return   % Exit 
        end

        try
          set_param([ModelName_Network,'/',Load_invName],'ActiveReactivePowers', sprintf('[%g %g ]',Pinv*1e6 , Qinv*1e6));
        catch
          msg_block = sprintf('HVDC Link: No block (PQ load) called ''%s'' could be found.',Load_invName);
          errordlg(msg_block,'Block Error');
          clear;
          load OriginalWorkSpace.mat;  % Load the original workspace
          return   % Exit 
        end
    else
        AC_DC_converge = 0;
        fprintf(fileID,'--> Does not converge after %d iterations\n',DC_max_iter);
        fprintf(fileID,'////////// Abnormal ending //////////\n');
       
        fprintf('--> Opened the report file: %s\n',ReportName);
        fprintf('////////// Abnormal ending //////////\n');
        open(ReportName);
        clear;
        load OriginalWorkSpace.mat;  % Load the original workspace
        return
    end     
    
    if (two_links)
    % === Link B ====
       sim(ModelName_LFDClink_B); % Execute the steady-state model of the second DC link
     
      if (i == 1)
        fprintf(fileID_B,'/// Sequential DC and AC load flows solution report: %s ///\rNetwork model: %s\n',...
        ReportName_B,ModelName_Network);
        fprintf(fileID_B,'DC link steady-state model: %s\n',ModelName_LFDClink_B);
        fprintf(fileID_B,'Date: %s\n\n',DateString);
        fprintf(fileID_B,'______ Execute the DC load flow (Iteration No %d) ______\n',i);
        fprintf(fileID_B,'Assumed values from AC network: \n');
      else
       fprintf(fileID_B,'______ Execute the DC load flow (Iteration No %d) ______\n',i);
       fprintf(fileID_B,'Inputs from AC load flow: \n');
      end
    
      fprintf(fileID_B,'Rectifier AC voltage: Eac_R = %5.2f kV\n', Eac_R_B(end));
      fprintf(fileID_B,'Inverter  AC voltage: Eac_I = %5.2f kV\n', Eac_I_B(end));
     
     if (DC_iter_B(end) <= DC_max_iter)

        fprintf(fileID_B,'--> Converges after %d iteration(s):\n',DC_iter_B(end));
        
        fprintf(fileID_B,'Rectifier: Tap = %5.4f (pu), alpha = %4.2f (deg), Udi0 = %5.2f (kV)\n',...
          Tap_pu_R_B(end), alpha_R_B(end), Udi0_R_B(end));
        fprintf(fileID_B,'Inverter:  Tap = %5.4f (pu), gamma = %4.2f (deg), Udi0 = %5.2f (kV)\n\n',...
          Tap_pu_I_B(end), gamma_I_B(end), Udi0_I_B(end));
   
        Prec_B=Pac_R_B(end);    % AC active power at rectifier (MW)
        Qrec_B=Qac_R_B(end);    % AC reactive power at rectifier (Mvar)
        Pinv_B=Pac_I_B(end);    % AC active power at inverter (MW)
        Qinv_B=Qac_I_B(end);    % AC reactive power at inverter (Mvar)
    
        % Write P, Q values, for the next AC load flow
        try
          set_param([ModelName_Network,'/',Load_recName_B],'ActiveReactivePowers', sprintf('[%g %g ]',Prec_B*1e6 , Qrec_B*1e6));
        catch
          msg_block_B = sprintf('Second HVDC Link: No block (PQ load) called ''%s'' could be found.',Load_recName_B);
          errordlg(msg_block_B,'Block Error');
          clear;
          load OriginalWorkSpace.mat;  % Load the original workspace
          return   % Exit 
        end
        
        try
          set_param([ModelName_Network,'/',Load_invName_B],'ActiveReactivePowers', sprintf('[%g %g ]',Pinv_B*1e6 , Qinv_B*1e6));
        catch
          msg_block_B = sprintf('Second DC Link: No block (PQ load) called ''%s'' could be found.',Load_invName_B);
          errordlg(msg_block_B,'Block Error');
          clear;
          load OriginalWorkSpace.mat;  % Load the original workspace
          return   % Exit 
        end
    else
        AC_DC_converge_B = 0;
        fprintf(fileID_B,'--> Does not converge after %d iterations\n',DC_max_iter);
        fprintf(fileID_B,'////////// Abnormal ending //////////\n');
       
        fprintf('--> Opened the report file: %s\n',ReportName);
        fprintf('////////// Abnormal ending //////////\n');
        open(ReportName_B);
        clear;
        load OriginalWorkSpace.mat;  % Load the original workspace
        return
     end  
    end
    
   %% Test for the simultaneous convergence of the AC and DC solutions
    
    if (TAPINI_R == Tap_pu_R(end) && TAPINI_I == Tap_pu_I(end) && AC_DC_converge && i > 1)
        Normal_end = 1;
    else
       Normal_end = 0;
    end
   
  if (two_links)
    if (TAPINI_R_B == Tap_pu_R_B(end) && TAPINI_I_B == Tap_pu_I_B(end) && AC_DC_converge_B && i > 1)
        Normal_end_B = 1;
    else
       Normal_end_B = 0;
    end
  end

    %   Print final results after normal ending 
    
 if (Normal_end && Normal_end_B)
        fprintf('--> Simultaneous convergence after %d AC load flow(s)\n',i-1);

        % === Link A
        fprintf(fileID,'--> Simultaneous convergence after %d DC load flow(s) and %d AC load flow(s)\n',i,i-1);
        fprintf(fileID,'Rectifier AC voltage:  %5.2f kV\n', Eac_R(end));
        fprintf(fileID,'Inverter  AC voltage:  %5.2f kV\n', Eac_I(end));
        fprintf(fileID,'Rectifier: Tap = %5.4f (pu), alpha = %4.2f (deg), Udi0 = %5.2f (kV)\n',...
          Tap_pu_R(end), alpha_R(end), Udi0_R(end));
        fprintf(fileID,'Inverter:  Tap = %5.4f (pu), gamma = %4.2f (deg), Udi0 = %5.2f (kV)\n\n',...
          Tap_pu_I(end), gamma_I(end), Udi0_I(end));

      
        fprintf(fileID,'Rectifier: Vdc = %5.2f (kV), Idc = %5.4f (kA), Pdc = %+6.2f (MW)\n',...
             Vdc_R(end), Idc_R(end), Pdc_R(end));
        fprintf(fileID,'Inverter:  Vdc = %5.2f (kV), alpha = %5.2f (deg), Pdc = %+6.2f (MW)\n\n',...
            Vdc_I(end), alpha_I(end), Pdc_I(end));
        fprintf(fileID,'Bus %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
            bus_recName,Prec, Qrec);
        fprintf(fileID,'Bus %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
            bus_invName,Pinv, Qinv);   
        
        if (Tap_limited_R(end))
            fprintf(fileID,'*** Warning --> Tap is limited at rectifier\n');
        end
        if(Tap_limited_I(end))
            fprintf(fileID,'*** Warning --> Tap is limited at inverter\n');
        end
        
        fprintf(fileID,'/////////// Normal ending //////////');
        fprintf('--> Opened the report file: %s\n',ReportName);
        open(ReportName);
        
  if (two_links)      
        % === Link B
        fprintf(fileID_B,'--> Simultaneous convergence after %d AC load flow(s)\n',i-1);
        fprintf(fileID_B,'Rectifier: Vdc = %5.2f (kV), Idc = %5.4f (kA), Pdc = %+6.2f (MW)\n',...
             Vdc_R_B(end), Idc_R_B(end), Pdc_R_B(end));
        fprintf(fileID_B,'Inverter:  Vdc = %5.2f (kV), alpha = %5.2f (deg), Pdc = %+6.2f (MW)\n\n',...
            Vdc_I_B(end), alpha_I_B(end), Pdc_I_B(end)); 
        fprintf(fileID_B,'Bus %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
            bus_recName_B,Prec_B, Qrec_B);
        fprintf(fileID_B,'Bus %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
            bus_invName_B,Pinv_B, Qinv_B);   
         if (Tap_limited_R_B)
            fprintf(fileID_B,'*** Warning --> Tap is limited at rectifier\n');
        end
        if(Tap_limited_I_B)
            fprintf(fileID_B,'*** Warning --> Tap is limited at inverter\n');
        end
        
        fprintf(fileID_B,'/////////// Normal ending //////////');
        fprintf('--> Opened the report file: %s\n',ReportName_B);
        open(ReportName_B);
  end
        
        %%   Set to zero, after the final solution, the PQ loads represeting the DC converters units
       % === Link A
       set_param([ModelName_Network,'/',Load_recName],'ActiveReactivePowers', sprintf('[%g %g ]',0 , 0));
       set_param([ModelName_Network,'/',Load_invName],'ActiveReactivePowers', sprintf('[%g %g ]',0 , 0)); 
      
   if (two_links)       
       % === Link B
       set_param([ModelName_Network,'/',Load_recName_B],'ActiveReactivePowers', sprintf('[%g %g ]',0 , 0));
       set_param([ModelName_Network,'/',Load_invName_B],'ActiveReactivePowers', sprintf('[%g %g ]',0 , 0));

   end
       
  % Close report file
   fclose(fileID);   % Link A
   if (two_links)
     fclose(fileID_B); % Link B
   end
  
  if (flag_clear)
   save outfile.mat Idc_R Tap_pu_R Tap_pu_I two_links;
   if (two_links)
     save outBfile.mat Idc_R_B Tap_pu_R_B Tap_pu_I_B;
   end
  
   clear;
    load OriginalWorkSpace.mat;  % Load the original workspace.
    load outfile.mat;       % Load desired output variables
   
    if (two_links)
      load outBfile.mat;       % Load desired output variables
    end
    
   end
    return      % Exit
 
 end

   
%% Determine the AC load flow solution of the AC network including the DC converters as PQ loads.
try
    LF = power_loadflow('-v2',ModelName_Network,'solve');
catch
    msg_block = sprintf('power_loadflow function generated an error message: check model for errors.');
    errordlg(msg_block,'AC load flow');
    clear;
    load OriginalWorkSpace.mat;  % Load the original workspace
    return   % Exit 
end

        if (LF.status == 1)               % Did converge   
      AC_iter = LF.iterations;        % Number of AC iterations
      AC_converge = 1;
    else
      AC_converge = 0;
    end
    
    
 % ----  Check for bus name identification error
    % === Link A
     if (i==1)        % The first time, defines the buses number
      ibus_rec = find(strcmp({LF.bus.ID},bus_recName));
      if (isempty(ibus_rec))
          msg1_R = sprintf('DC link: Rectifier''s load flow bus name %s not found.',bus_recName);
          errordlg(msg1_R,'Bus Error');
          clear;
          load OriginalWorkSpace.mat;
          return    % exit
      end
      
      ibus_inv = find(strcmp({LF.bus.ID},bus_invName));
       if (isempty(ibus_inv))
          msg1_I = sprintf('DC link: Inverter''s load flow bus name %s not found.',bus_invName);
          errordlg(msg1_I,'Bus Error');
          clear;
          load OriginalWorkSpace.mat;
          return    % Exit
      end
     end
     
    if (two_links)
     % === Link B
     if (i==1)        % The first time
      ibus_rec_B = find(strncmp({LF.bus.ID},bus_recName_B,4));
      if (isempty(ibus_rec_B))
          msg1_R = sprintf('Second DC link: Rectifier''s load flow bus name %s not found.',bus_recName);
          errordlg(msg1_R,'Bus Error');
          clear;
          load OriginalWorkSpace.mat;
      end
      
      ibus_inv_B = find(strncmp({LF.bus.ID},bus_invName_B,4));
      if (isempty(ibus_inv_B))
          msg1_I = sprintf('Second DC link: Inverter''s load flow bus name %s not found.',bus_invName);
          errordlg(msg1_I,'Bus Error');
          clear;
          load OriginalWorkSpace.mat;
      end
     end
    end
     
    % == Link A : Print 
    
    fprintf(fileID,'~~~~~~ Execute the AC load flow (Iteration No %d) ~~~~~~\n',i);
    fprintf(fileID,'Inputs from the DC load flow: \n');
    fprintf(fileID,'Rectifier at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_recName, Prec, Qrec);
    fprintf(fileID,'Inverter  at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_invName, Pinv, Qinv);   
    
    if (AC_converge)
         fprintf(fileID,'--> Converges after %d iteration(s)\n',AC_iter);
         fprintf(fileID,'Bus %s: V = %4.3f pu %6.2f deg\n',...
             bus_recName, abs(LF.bus(ibus_rec).Vbus),angle(LF.bus(ibus_rec).Vbus)*180/pi);
         fprintf(fileID,'Bus %s: V = %4.3f pu %6.2f deg\n\n',...
             bus_invName, abs(LF.bus(ibus_inv).Vbus),angle(LF.bus(ibus_inv).Vbus)*180/pi);
    else
      fprintf(fileID,['--> ',LF.error,'\n']);
      fprintf(fileID,'////////// Abnormal ending //////////\n');
      open(ReportName);
    end
    
  if (two_links)
    % == Link B : Print 
    
    fprintf(fileID_B,'~~~~~~ Execute the AC load flow (Iteration No %d) ~~~~~~\n',i);
    fprintf(fileID_B,'Inputs from the DC load flow: \n');
    fprintf(fileID_B,'Rectifier at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_recName_B, Prec_B, Qrec_B);
    fprintf(fileID_B,'Inverter  at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_invName_B, Pinv_B, Qinv_B);   
    
    if (AC_converge)
         fprintf(fileID_B,'--> Converges after %d iteration(s)\n',AC_iter);
         fprintf(fileID_B,'Bus %s: V = %4.3f pu %6.2f deg\n',...
             bus_recName, abs(LF.bus(ibus_rec).Vbus),angle(LF.bus(ibus_rec).Vbus)*180/pi);
         fprintf(fileID_B,'Bus %s: V = %4.3f pu %6.2f deg\n\n',...
             bus_invName, abs(LF.bus(ibus_inv).Vbus),angle(LF.bus(ibus_inv).Vbus)*180/pi);
    else
      fprintf(fileID_B,['--> ',LF.error,'\n']);
      fprintf(fileID_B,'////////// Abnormal ending //////////\n');
      open(ReportName_B);
    end
  end
    
    if (~AC_converge)
         fprintf('////////// Abnormal ending //////////\n');
         fprintf('--> Opened the report file: %s\n',ReportName);
         return
    end
    
     % After AC convergence, write variables, for the next DC load flow
     
     % === Link A
     Eac_R_INI    = abs(LF.bus(ibus_rec).Vbus) * UacNom_R;  % (kV)
     Eac_I_INI    = abs(LF.bus(ibus_inv).Vbus) * UacNom_I; 
     set_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'Eac_R_INI',[num2str(Eac_R_INI), '  %Updated']);
     set_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'Eac_I_INI',[num2str(Eac_I_INI), '  %Updated']);
     TAPINI_R = Tap_pu_R(end);                     % (pu)
     TAPINI_I = Tap_pu_I(end); 
     set_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'TAP_R_INI',[num2str(TAPINI_R), '  %Updated']);
     set_param([ModelName_LFDClink,'/HVDC Link Steady-State'],'TAP_I_INI',[num2str(TAPINI_I), '  %Updated']);
     
   if (two_links)
     % === Link B   
     Eac_R_INI    = abs(LF.bus(ibus_rec_B).Vbus) * UacNom_R_B;  % (kV)
     Eac_I_INI    = abs(LF.bus(ibus_inv_B).Vbus) * UacNom_I_B;  
     set_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'Eac_R_INI',[num2str(Eac_R_INI), '  %Updated']);
     set_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'Eac_I_INI',[num2str(Eac_I_INI), '  %Updated']);
     TAPINI_R_B = Tap_pu_R_B(end);                     % (pu)
     TAPINI_I_B = Tap_pu_I_B(end); 
     set_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'TAP_R_INI',[num2str(TAPINI_R_B), '  %Updated']);
     set_param([ModelName_LFDClink_B,'/Second HVDC Link Steady-State'],'TAP_I_INI',[num2str(TAPINI_I_B), '  %Updated']);
   end
     
 
   end

%   Print final results after abnormal ending 
  
% === Link A
fprintf(fileID,'AC and DC load flows have not simulateously converged after %d DC load flows\n',DC_AC_max);
fprintf(fileID,'Rectifier: Vdc = %5.2f (kV), Idc = %5.4f (kA), Pdc = %+6.2f (MW)\n',...
        Vdc_R(end), Idc_R(end), Pdc_R(end));
fprintf(fileID,'Inverter:  Vdc = %5.2f (kV), alpha = %5.2f (deg), Pdc = %+6.2f (MW)\n\n',...
        Vdc_I(end), alpha_I(end), Pdc_I(end)); 
fprintf(fileID,'Rectifier at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_recName, Prec, Qrec);
fprintf(fileID,'Inverter  at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_invName, Pinv, Qinv); 
fprintf(fileID,'////////// Abnormal ending //////////\n');

if (two_links)
% === Link B
fprintf(fileID_B,'AC and DC load flows have not simulateously converged after %d DC load flows\n',DC_AC_max);
fprintf(fileID_B,'Rectifier: Vdc = %5.2f (kV), Idc = %5.4f (kA), Pdc = %+6.2f (MW)\n',...
        Vdc_R_B(end), Idc_R_B(end), Pdc_R_B(end));
fprintf(fileID_B,'Inverter:  Vdc = %5.2f (kV), alpha = %5.2f (deg), Pdc = %+6.2f (MW)\n\n',...
        Vdc_I_B(end), alpha_I_B(end), Pdc_I_B(end)); 
fprintf(fileID_B,'Rectifier at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_recName, Prec, Qrec);
fprintf(fileID_B,'Inverter  at %s: P = %+7.2f MW, Q = %7.2f Mvar\n',...
          bus_invName, Pinv, Qinv); 
fprintf(fileID_B,'////////// Abnormal ending //////////\n');
% 
end

disp('////////// Abnormal ending //////////');

% Close report file
fclose(fileID);   % Link A

if (two_links)
fclose(fileID_B); % Link B
end

clear;
load OriginalWorkSpace.mat;

% //////////////////////// E N D \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
